@@ -624,13 +624,20 @@ class client():
         resp.raise_for_status()
         return resp.json()
 
-    def register_sdf(self, file, userName, mappings):
+    def register_sdf(self, file, userName, mappings, prefix = None):
         files = self.upload_files([file])
         request = {
             "fileName": files['files'][0]["name"],
             "userName": userName,
-            "mappings": mappings
+            "mappings": mappings,
         }
+        if prefix:
+            request["labelPrefix"] = {
+                "name": prefix,
+                "labelPrefix": prefix,
+                "labelTypeAndKind":"id_corpName",
+                "thingTypeAndKind":"parent_compound"
+            }
         response = self.register_sdf_request(request)
         report_files = []
         for file in response[0]['reportFiles']:
@@ -792,7 +799,7 @@ class client():
             return None
         return resp.json()
 
-    def get_ls_thing(self, ls_type, ls_kind, code_name, nestedfull=False):
+    def get_ls_thing(self, ls_type, ls_kind, code_name, nestedfull=True):
         """
         Get a models.LsThing object by ls_type, ls_kind, and code_name
 
@@ -834,7 +841,7 @@ class client():
         return resp.json()
 
     def get_ls_things_by_codes(self, ls_type, ls_kind, code_name_list,
-                               nestedfull=False):
+                               nestedfull=True):
         """
         Get a list of ls thing dict objects from a list of their code_names
 
@@ -843,11 +850,14 @@ class client():
             ls_kind (str): ls_kind for all things to retrieve
             code_name_list (str list): list of str code_names
         """
+        params = {}
+        if nestedfull:
+            params = {**params, 'with': 'nestedfull'}
         resp = self.session.post("{}/api/things/{}/{}/codeNames/jsonArray".
                                  format(self.url,
                                         ls_type,
                                         ls_kind),
-                                 params={'nestedfull': nestedfull},
+                                 params=params,
                                  headers={'Content-Type': "application/json"},
                                  data=json.dumps(code_name_list))
         if resp.status_code == 500:
@@ -856,6 +866,31 @@ class client():
             resp.raise_for_status()
         return resp.json()
 
+    def get_ls_things_by_type_and_kind(self, ls_type, ls_kind,
+                               format='stub'):
+        """
+        Get a list of ls thing dict objects from ls_type and ls_kind
+
+        Args:
+            ls_type (str): ls_type for all things to retrieve
+            ls_kind (str): ls_kind for all things to retrieve
+        """
+        allowedFormats = {'codetable', 'stub'}
+        if format not in allowedFormats:
+            raise ValueError("format must be one of %s." % allowedFormats)
+        params = {format: '1'}
+        resp = self.session.get("{}/api/things/{}/{}".
+                                 format(self.url,
+                                        ls_type,
+                                        ls_kind),
+                                 params=params,
+                                 headers={'Content-Type': "application/json"})
+        if resp.status_code == 500:
+            return None
+        else:
+            resp.raise_for_status()
+        return resp.json()
+
     def save_ls_thing_list(self, ls_thing_list):
         """
         Save a list of ls thing dict objects
@@ -885,14 +920,16 @@ class client():
         resp.raise_for_status()
         return resp.json()
 
-    def get_thing_codes_by_labels(self, ls_type, ls_kind, label_list):
+    def get_thing_codes_by_labels(self, thing_type, thing_kind, labels_or_codes, label_type=None, label_kind=None):
         """
         Get a list of thing codes by providing a list of labels
 
         Args:
-            ls_type (str): ls_type for all things to retrieve
-            ls_kind (str): ls_kind for all things to retrieve
-            label_list (str list): list of str labels
+            labels_or_codes (str list): list of str labels or codes
+            thing_type (str): ls_type for all things to retrieve
+            thing_kind (str): ls_kind for all things to retrieve
+            label_type (str): label_type to limit label searches
+            label_kind (str): label_kind to limit label searches
         Returns:
             ref_name_lookup_results: list of objects with
                 requestName (str): input label string
@@ -900,17 +937,17 @@ class client():
                 referenceName (str): LsThing code name string
         """
         request = {
-            'thingType': ls_type,
-            'thingKind': ls_kind,
+            'thingType': thing_type,
+            'thingKind': thing_kind,
+            'labelType': label_type,
+            'labelKind': label_kind,
             'requests': [
-                {"requestName": label} for label in label_list
-                ]
+                    {"requestName": request} for request in labels_or_codes
+            ]
         }
 
-        resp = self.session.post("{}/api/getThingCodeByLabel/{}/{}".
-                                 format(self.url,
-                                        ls_type,
-                                        ls_kind),
+        resp = self.session.post("{}/api/getThingCodeByLabel".
+                                 format(self.url),
                                  headers={'Content-Type': "application/json"},
                                  data=json.dumps(request))
         if resp.status_code == 500:
@@ -918,6 +955,8 @@ class client():
         else:
             resp.raise_for_status()
         resp_object = resp.json()
+        if resp_object == 'error trying to lookup lsThing name':
+            raise RuntimeError("Failed to get things, please see acas logs.")
         results = resp_object['results']
         return results
 
@@ -999,3 +1038,137 @@ class client():
             return [res['code'] for res in results]
         else:
             return results
+
+    def create_label_sequence(self, labelPrefix, startingNumber, digits,
+                                  labelSeparator, labelTypeAndKind = None, thingTypeAndKind = None,
+                                  labelSequenceRoles =  []):
+        """
+        Create a label sequence
+
+        Args:
+            labelPrefix (str): Prefix of the label
+            startingNumber (str): Set to 0 for the first number to be 1
+            digits (str): The number of leading zeros to add to the label sequence when formatting (e.g. CMPD-0000007 would be digts: 7 )
+            labelTypeAndKind (str): the label type and kind associated with this sequence (used for finding all labels of a specific label type and kind in some interfaces)
+            thingTypeAndKind (str): the thing type and kind associated with this sequence (used for finding all labels of a specific thing type and kind in some interfaces)
+            labelSequenceRoles (list): the registered role to associate with this label (used for limiting access to specific label sequences in some interfaces)
+
+        Returns:
+            a dict object representing the new label sequence
+        """
+        request = {
+            'labelPrefix': labelPrefix,
+            'startingNumber': startingNumber,
+            'digits': digits,
+            'labelSeparator': labelSeparator,
+            'labelTypeAndKind': labelTypeAndKind,
+            'thingTypeAndKind': thingTypeAndKind,
+            'labelSequenceRoles': labelSequenceRoles
+        }
+        resp = self.session.post("{}/api/labelsequences/".
+                                 format(self.url),
+                                 headers={'Content-Type': "application/json"},
+                                 data=json.dumps(request))
+        resp.raise_for_status()
+        return resp.json()
+
+    def get_all_label_sequences(self):
+        """
+        Get all label sequences (limited to those authorized by logged in user roles)
+
+        Returns:
+            a list of dict objects representing the labelSequence
+        """
+       
+        json = self.get_label_sequence_by_types_and_kinds()
+        return json
+
+    def get_label_sequence_by_types_and_kinds(self, labelTypeAndKind = None, thingTypeAndKind = None):
+        """
+        Get label sequence by types and kinds (limited to those authorized by logged in user roles)
+
+        Args:
+            labelPrefix (str): Prefix of the label
+            startingNumber (str): Set to 0 for the first number to be 1
+            digits (str): The number of leading zeros to add to the label sequence when formatting (e.g. CMPD-0000007 would be digts: 7 )
+            labelTypeAndKind (str): the label type and kind associated with this sequence (used for finding all labels of a specific label type and kind in some interfaces)
+            thingTypeAndKind (str): the thing type and kind associated with this sequence (used for finding all labels of a specific thing type and kind in some interfaces)
+            labelSequenceRoles (list): the registered role to associate with this label (used for limiting access to specific label sequences in some interfaces)
+
+        Returns:
+            a list of dict objects representing the labelSequence
+        """
+        params = {}
+        if labelTypeAndKind:
+            params = {**params, 'labelTypeAndKind': labelTypeAndKind}
+        if thingTypeAndKind:
+            params = {**params, 'thingTypeAndKind': thingTypeAndKind}
+        resp = self.session.get("{}/api/labelSequences/getAuthorizedLabelSequences".
+                                 format(self.url),
+                                 params=params)
+        resp.raise_for_status()
+        return resp.json()
+
+    def get_labels(self, labelTypeAndKind, thingTypeAndKind, numberOfLabels):
+        """
+        Get next n labels from label sequence prefix
+
+        Args:
+            labelTypeAndKind (str): Prefix of the registered label (see create_label_sequence)
+            numberOfLabels (int): Number of labels to fetch
+
+        Returns:
+            a list of dict objects representing the labelSequence
+        """
+        request = {
+            'labelTypeAndKind': labelTypeAndKind,
+            'thingTypeAndKind': thingTypeAndKind,
+            'numberOfLabels': numberOfLabels
+        }
+        resp = self.session.post("{}/api/getNextLabelSequence".
+                                 format(self.url),
+                                 headers={'Content-Type': "application/json"},
+                                 data=json.dumps(request))
+        resp.raise_for_status()
+        return resp.json()
+
+    def get_all_ddict_values(self):
+        """
+        Get all ddict values
+
+        Returns:
+            a list of dict objects representing the ddict value (aka code value)
+        """
+        all_values = self.get_ddict_values_by_type_and_kind()
+        return all_values
+
+    def get_ddict_values_by_type_and_kind(self, codeType=None, codeKind=None):
+        """
+        Get ddict values
+
+        Returns:
+            a list of dict objects representing the ddict value (aka code value)
+        """
+        path = "/api/codetables"
+        if codeType and codeKind:
+            path = "{}/{}/{}".format(path, codeType, codeKind)
+        resp = self.session.get("{}{}".
+                                format(self.url, path))
+        resp.raise_for_status()
+        return resp.json()
+
+
+    def get_blob_data_by_value_id(self, valueId):
+        """
+        Get blob data by value id
+        Args:
+            valueId (int): A known value id to fetch from the database that is stored as a blobValue lsType
+
+        Returns:
+            (bytes): representing the blob value
+        """
+
+        resp = self.session.get("{}/api/thingvalues/downloadThingBlobValueByID/{}"
+                                .format(self.url, valueId))
+        resp.raise_for_status()
+        return resp.content
