@@ -14,7 +14,11 @@ import uuid
 # "PROJ-00000001" registered
 
 
-def create_project_thing(code):
+def create_project_thing(code, name=None, alias=None):
+    if name is None:
+        name = code
+    if alias is None:
+        alias = name
     ls_thing = {
             "lsType": "project",
             "lsKind": "project",
@@ -24,7 +28,7 @@ def create_project_thing(code):
                 {
                     "lsType": "name",
                     "lsKind": "project name",
-                    "labelText": code,
+                    "labelText": name,
                     "ignored": False,
                     "preferred": True,
                     "recordedDate": 1586877284571,
@@ -36,7 +40,7 @@ def create_project_thing(code):
                 {
                     "lsType": "name",
                     "lsKind": "project alias",
-                    "labelText": code,
+                    "labelText": alias,
                     "ignored": False,
                     "preferred": False,
                     "recordedDate": 1586877284571,
@@ -90,6 +94,40 @@ def create_project_thing(code):
         }
     return ls_thing
 
+
+def create_thing_with_blob_value(code):
+    # Function for creating a thing with a blob value
+    # Returns a  thing, file name and bytes_array for unit testing purposes
+
+    # Get a file to load
+    file_name = 'blob_test.png'
+    blob_test_path = Path(__file__).resolve().parent\
+            .joinpath('test_acasclient', file_name)
+    f = open(blob_test_path, "rb")
+    bytes_array = f.read()
+
+    # Need to save blob value as an int array not bytes
+    int_array_to_save = [x for x in bytes_array]
+    f.close()
+
+    # Create an Ls thing and add the blob value
+    # comments should be the file name
+    code = str(uuid.uuid4())
+    ls_thing = create_project_thing(code)
+    blob_value = {
+        "lsType": "blobValue",
+        "blobValue": int_array_to_save,
+        "lsKind": "my file",
+        "ignored": False,
+        "recordedDate": 1586877284571,
+        "recordedBy": "bob",
+        "comments": file_name
+    }
+    ls_thing["lsStates"][0]["lsValues"].append(blob_value)
+
+    # Return thing file and bytes array for testing
+    return ls_thing, file_name, bytes_array
+
 class TestAcasclient(unittest.TestCase):
     """Tests for `acasclient` package."""
 
@@ -942,18 +980,81 @@ class TestAcasclient(unittest.TestCase):
         self.assertIn('codeName', updated_ls_things[0])
 
     def test_030_get_thing_codes_by_labels(self):
-        labels = []
+        codes = []
+        names = []
+        aliases = []
         for n in range(3):
-            label = str(uuid.uuid4())
-            ls_thing = create_project_thing(label)
+            code = str(uuid.uuid4())
+            name = str(uuid.uuid4())
+            alias = str(uuid.uuid4())
+            ls_thing = create_project_thing(code, name, alias)
             self.client.save_ls_thing(ls_thing)
-            labels.append(label)
+            codes.append(code)
+            names.append(name)
+            aliases.append(alias)
 
+        # Verify search by code type and kind works without label filter
         results = self.client.get_thing_codes_by_labels('project',
                                                         'project',
-                                                        labels)
-        self.assertEqual(len(results), len(labels))
+                                                        codes)
+
+        self.assertEqual(len(results), len(codes))
         self.assertIn('preferredName', results[0])
+        # Preferred name should be the name sent to the service as preferred, not the code
+        self.assertIn(results[0]["preferredName"], names)
+
+        # Adding label type and label kind should not stop searching by code name
+        # but the preferred ids should still be the names
+        results = self.client.get_thing_codes_by_labels('project',
+                                                        'project',
+                                                        codes,
+                                                        'garbageshouldnotexist',
+                                                        'garbageshouldnotexist')
+        self.assertEqual(len(results), len(codes))
+        self.assertIn(results[0]["preferredName"], names)
+
+        # Searching by labels without limiting by label type/kind  should give results
+        results = self.client.get_thing_codes_by_labels('project',
+                                                        'project',
+                                                        names)
+        self.assertEqual(len(results), len(names))
+        self.assertIn(results[0]["preferredName"], names)
+
+        # Searching by names and filtering label types and kinds should give no results when no matching
+        results = self.client.get_thing_codes_by_labels('project',
+                                                        'project',
+                                                        names,
+                                                        'garbageshouldnotexist',
+                                                        'garbageshouldnotexist')
+        self.assertEqual(len(results), len(names))
+        self.assertEqual(results[0]["preferredName"], '')
+
+        # Searching when code and label is the same should still produce a single response
+        codeAndName = str(uuid.uuid4())
+        alias = str(uuid.uuid4())
+        ls_thing = create_project_thing(codeAndName, codeAndName, alias)
+        self.client.save_ls_thing(ls_thing)
+        results = self.client.get_thing_codes_by_labels('project',
+                                              'project',
+                                              [codeAndName])
+        self.assertEqual(results[0]["preferredName"], codeAndName)
+
+        # Searching by alias should work but still return preferredNames
+        results = self.client.get_thing_codes_by_labels('project',
+                                                        'project',
+                                                        aliases)
+        self.assertEqual(len(results), len(names))
+        self.assertIn(results[0]["preferredName"], names)
+
+        # Searching by alias should work but still return preferredNames even when specifying labeltype and kind
+        results = self.client.get_thing_codes_by_labels('project',
+                                                        'project',
+                                                        aliases,
+                                                        'name',
+                                                        'project alias')
+        self.assertEqual(len(results), len(names))
+        self.assertIn(results[0]["preferredName"], names)
+
 
     def test_031_get_saved_entity_codes(self):
         labels = []
@@ -1010,3 +1111,110 @@ class TestAcasclient(unittest.TestCase):
         self.assertIn('parentCorpName', all_lots[0])
         self.assertIn('registrationDate', all_lots[0])
         self.assertIn('project', all_lots[0])
+
+    def test_034_get_ls_things_by_type_and_kind(self):
+        codes = []
+        for n in range(3):
+            code = str(uuid.uuid4())
+            ls_thing = create_project_thing(code)
+            self.client.save_ls_thing(ls_thing)
+            codes.append(ls_thing["codeName"])
+
+        ls_things = self.client.get_ls_things_by_type_and_kind("project",
+                                                               "project",
+                                                               "stub")
+        self.assertIn('codeName', ls_things[0])
+        returnedCodes = []
+        for thing in ls_things:
+            if thing['codeName'] in codes:
+                returnedCodes.append(thing['codeName'])
+        self.assertEqual(len(returnedCodes), len(codes))
+        self.assertIn('codeName', ls_things[0])
+
+        #Verify that codectable format works
+        ls_things = self.client.get_ls_things_by_type_and_kind("project",
+                                                               "project",
+                                                               format="codetable")
+        self.assertIn('code', ls_things[0])
+
+        #Verify that giving bad format gives ValueError
+        with self.assertRaises(ValueError):
+            self.client.get_ls_things_by_type_and_kind("project",
+                                                       "project",
+                                                       format="badformat")                                 
+
+    def test_035_test_create_label_sequence(self):
+        labelPrefix = "TESTSEQ"+str(uuid.uuid4())
+        sequence = self.client.create_label_sequence(labelPrefix, 0, 0, "-", "id_corpName", "parent_compound")
+        self.assertIn('dbSequence', sequence)
+        self.assertIn(labelPrefix, sequence["labelPrefix"])
+        self.assertIn('id_corpName', sequence["labelTypeAndKind"])
+        self.assertIn('parent_compound', sequence["thingTypeAndKind"])
+
+
+    def test_036_get_all_label_sequences(self):
+        sequences = self.client.get_all_label_sequences()
+        self.assertGreater(len(sequences), 0)
+        self.assertIn('dbSequence', sequences[0])
+        self.assertIn('labelPrefix', sequences[0])
+        self.assertIn('thingTypeAndKind', sequences[0])
+
+    def test_037_get_label_sequence_by_types_and_kinds(self):
+        labelTypeAndKind = "id_corpName"
+        thingTypeAndKind = "parent_compound"
+        sequences = self.client.get_label_sequence_by_types_and_kinds(labelTypeAndKind, thingTypeAndKind)
+        self.assertGreater(len(sequences), 0)
+        for sequence in sequences:
+            self.assertEqual(labelTypeAndKind, sequence["labelTypeAndKind"])
+            self.assertEqual(thingTypeAndKind, sequence["thingTypeAndKind"])
+
+    def test_038_get_labels(self):
+        numberOfLabels = 5
+        labels = self.client.get_labels("id_codeName", "document_experiment", numberOfLabels)
+        self.assertEqual(len(labels), numberOfLabels)
+        for label in labels:
+            self.assertIn('autoLabel', label)
+
+    def test_039_get_all_ddict_values(self):
+        all_ddict_values = self.client.get_all_ddict_values()
+        self.assertGreater(len(all_ddict_values), 0)
+        for ddict_value in all_ddict_values:
+            self.assertIn('codeType', ddict_value)
+            self.assertIn('codeKind', ddict_value)
+            self.assertIn('code', ddict_value)
+
+    def test_040_get_ddict_values_by_type_and_kind(self):
+        codeType = "experiment metadata"
+        codeKind = "file type"
+        all_ddict_values = self.client.get_ddict_values_by_type_and_kind(codeType, codeKind)
+        self.assertGreater(len(all_ddict_values), 0)
+        for ddict_value in all_ddict_values:
+            self.assertIn('codeType', ddict_value)
+            self.assertIn('codeKind', ddict_value)
+            self.assertEqual(codeType, ddict_value["codeType"])
+            self.assertEqual(codeKind, ddict_value["codeKind"])
+
+    def test_041_get_blob_data_by_value_id(self):
+        # Save an ls thing with a blob value
+        code = str(uuid.uuid4())
+        ls_thing, file_name, bytes_array = create_thing_with_blob_value(code)
+        saved_ls_thing = self.client.save_ls_thing(ls_thing)
+
+        # Get the blob value from the saved ls thing (does not contain blobValue data)
+        saved_blob_value = None
+        for state in saved_ls_thing["lsStates"]:
+            for value in state["lsValues"]:
+                if value["lsType"] == "blobValue":
+                    saved_blob_value = value
+                    break
+
+        # Blob value should return and the comments should be set to the file name
+        self.assertIsNotNone(saved_blob_value)
+        self.assertEqual(saved_blob_value["comments"], file_name)
+
+        # Get the actual blob value data by value id
+        blob_data = self.client.get_blob_data_by_value_id(saved_blob_value["id"])
+
+        # Assert that the returned blob data is of type bytes and is equal to the blob data sent in
+        self.assertEqual(type(blob_data), bytes)
+        self.assertEqual(blob_data, bytes_array)
